#include <iostream>
#include <cassert>
#include <set>
#include <utility>
#include <vector>
#include <map>
#include <algorithm>

using namespace std;

#define forn(i, n) for (int i = 0; i < int(n); ++i)

#define ll long long
#define mp make_pair
#define pb push_back
#define all(v) v.begin(), v.end()

#define LOOSE_PERCENT 40
#define WIN_PERCENT 30

ll rand64(int k) {
	ll res = 0;
	forn (i, 4) res = (res << 16) ^ rand();
	return res & ((1LL << k) - 1);
}

void getPairSubtree(ll vertex, ll t, ll k, ll& left, ll& right) {
	vertex = (1LL << k) + vertex;
	vertex >>= (t - 1);
	vertex ^= 1;
	left = vertex << (t - 1);
	right = left | ((1LL << (t - 1)) - 1);

	left -= 1LL << k;
	right -= 1LL << k;
}

map<ll, int> looseInRound;

set<pair<ll, ll> > matches;

int _n;

void makeLooser(ll v, ll t, ll k) {

	if (looseInRound.count(v)) {
		if (looseInRound[v] <= t) return;
	}

	ll left, right;
	getPairSubtree(v, t, k, left, right);		
	
	if (t == 1) {
		assert(left == right);				
		if ((int)matches.size() < _n && matches.count(mp(v, left)) == 0) matches.insert(mp(left, v));
		return;
	}

	vector<ll> win;
	vector<ll> loose;
	vector<ll> notplay;

	for (ll i = left; i <= right; ++i) {
		int event = rand() % 10;
		if (event < LOOSE_PERCENT && matches.count(mp(v, i)) == 0) {
			loose.pb(i);																				
		} else
		if (event >= 100 - WIN_PERCENT && matches.count(mp(i, v)) == 0) {
			win.pb(i);
		} else {
			notplay.pb(i);
		}
	}
	random_shuffle(all(loose));
	random_shuffle(all(win));
	random_shuffle(all(notplay));

	if (loose.size() == 0) {
		if (notplay.size() != 0) {
			loose.pb(notplay.back());
			notplay.pop_back();
		} else {
			loose.pb(win.back());
			win.pop_back();
		}
	}

	forn (i, loose.size()) {
		if ((int)matches.size() < _n && matches.count(mp(v, loose[i])) == 0) matches.insert(mp(loose[i], v));		
		if (rand() % 100 <= 50) makeLooser(loose[i], rand() % (t - 1) + 1, k);
	}

	forn (i, win.size()) {
		if ((int)matches.size() < _n && matches.count(mp(win[i], v)) == 0) matches.insert(mp(v, win[i]));
		makeLooser(win[i], rand() % (t - 1) + 1, k);
	}

	forn (i, notplay.size()) {
		makeLooser(notplay[i], rand() % (t - 1) + 1, k);
	}

	if (looseInRound.count(v)) {
		looseInRound[v] = min(looseInRound[v], (int)t);
	} else {
		looseInRound[v] = (int)t;
	}
}
                  
int main(int argc, char** argv) {
	if (argc <= 4) {
		cout << "Usage: gen_random.exe <k> <n> <m> <seed>" << endl;
		return 1;		            	
	}	

	int k = atoi(argv[1]);
	int n = atoi(argv[2]);
	int m = atoi(argv[3]);
	int seed = atoi(argv[4]);
	srand(seed);

	assert(1 <= k && k <= 60);	
	assert(0 <= n && n <= 100000);	
	assert(1 <= m && m <= 100000);	

	_n = n;

	int max_tour = 1;
	while ((1LL << (max_tour)) <= n) ++max_tour;

	set<ll> kill;
	int it = 0;
	while (matches.size() < n - 2 && it <= 100000) {
		++it;
		ll v = rand64(k);		
		if (looseInRound.count(v)) continue;
		int tour = rand() % (min(k, max_tour) - 1) + 1;

		cerr << v << " " << tour << endl;

		kill.insert(v);
		makeLooser(v, tour, k);
	}

	while (kill.size() < m) kill.insert(rand64(k));

	vector<pair<ll, ll> > result(all(matches));
	random_shuffle(all(result));
	
	cout << k << " " << result.size() << " " << kill.size() << endl;

	forn (i, result.size()) cout << result[i].first + 1 << " " << result[i].second + 1 << endl;
	for (set<ll>::iterator it = kill.begin(); it != kill.end(); ++it) {
		if (it != kill.begin()) cout << " ";
		cout << *it + 1;						
	}		
	cout << endl;

	return 0;
}