#include <fstream>
#include <iostream>
#include <cstdio>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
using namespace std;

#define int64 long long

const int MAXK = 60;
const int MAXN = 100000;

class Edge
{
public:
    int64 be, en;
};

int n, m, k;
vector<Edge> edg[MAXK + 1];

int GetRand()
{
    int a = rand();
    a <<= 16;
    a |= rand();
    if (a < 0) a = -a;
    return a;
}

int GetTour(int64 a, int64 b)
{
//  cerr << "GetTour(" << a << "," << b << ") = ";
    a--;
    b--;
    int cur = 0;
    int64 cpow = 1;
    while (a != b)
    {
        a &= (~cpow);
        b &= (~cpow);
        cur++;
        cpow <<= 1;
    }
//  cerr << cur << "\n";
    return cur;
}

vector<int64> res;

void Load()
{
    res.clear();
    scanf("%d%d%d", &k, &n, &m);
    int i;
    for (i = 0; i <= k; i++) edg[i].clear();
    for (i = 0; i < n; i++)
    {
        int64 p, q;
        scanf("%I64d%I64d", &p, &q);
        Edge a;
        a.be = p;
        a.en = q;
        edg[GetTour(p, q)].push_back(a);
    }
    res.clear();
    for (i = 0; i < m; i++)
    {
        int64 p;
        scanf("%I64d", &p);
        res.push_back(p);
    }
}

int64 killed[MAXN];
int numkilled;
map<int64, int> maxtour;
int first[2 * MAXN];
int next[2 * MAXN];
int rfirst[2 * MAXN];
int rnext[2 * MAXN];
map<int64, int> pos;

int GetPos(int64 k)
{
    if (pos.find(k) == pos.end())
    {
        int val = pos.size();
        pos[k] = val;
    }
    return pos[k];
}

void CountFirstNext(int tnum)
{
    memset(first, 0xFF, sizeof(first));
    memset(next, 0xFF, sizeof(next));
    memset(rfirst, 0xFF, sizeof(rfirst));
    memset(rnext, 0xFF, sizeof(rnext));
    pos.clear();
    int i;
    for (i = 0; i < edg[tnum].size(); i++)
    {
        int tpos = GetPos(edg[tnum][i].be);
        next[i] = first[tpos];
        first[tpos] = i;
        tpos = GetPos(edg[tnum][i].en);
        rnext[i] = rfirst[tpos];
        rfirst[tpos] = i;
    }
}

int FindMaxLE(int64 val)
{
    int l = 0;
    int r = numkilled - 1;
    int max = -1;
    while (l <= r)
    {
        int mid = (l + r) >> 1;
        if (killed[mid] <= val)
        {
            if (mid > max) max = mid;
            l = mid + 1;
        }               
        else r = mid - 1;
    }
    return max;
}

int FindMinGE(int64 val)
{
    int l = 0;
    int r = numkilled - 1;
    int min = r + 1;
    while (l <= r)
    {
        int mid = (l + r) >> 1;
        if (killed[mid] >= val)
        {
            if (mid < min) min = mid;
            r = mid - 1;
        }
        else l = mid + 1;
    }
    return min;
}

int64 GetKilled(int64 lb, int64 rb)
{
    int ll, rr;
    ll = FindMinGE(lb);
    rr = FindMaxLE(rb);
    if (ll < 0) return 0;
    if (rr < 0) return 0;
    if (ll > rr) return 0;
    return (rr - ll + 1);
}

int Find(int64 cur)
{
    int l = 0;
    int r = numkilled - 1;
    while (l <= r)
    {
        int mid = (l + r) >> 1;
        if (killed[mid] == cur) return 1;
        if (killed[mid] < cur) l = mid + 1;
        else r = mid - 1;
    }
    return 0;
}

void QSort(int be, int en)
{
    int l = be;
    int r = en;
    int x = killed[GetRand() % (r - l + 1) + l];
    int t;
    while (l < r)
    {
        while (killed[l] < x) l++;
        while (killed[r] > x) r--;
        if (l <= r)
        {
            t = killed[l];
            killed[l] = killed[r];
            killed[r] = t;
            l++;
            r--;
        }
    }
    if (be < r) QSort(be, r);
    if (l < en) QSort(l, en);
}

void Solve()
{
    maxtour.clear();
    int i;
    int64 numunder = 2;
    numkilled = 0;
    for (i = 1; i <= k; i++)
    {
        //cerr << "Current tour " << i << "\n";
        if (numunder / 2 > n) break;
        int newnumkilled = numkilled;
        CountFirstNext(i);
        vector<int64> cur;
        cur.clear();
        int j;
        for (j = 0; j < edg[i].size(); j++)
        {
            cur.push_back(edg[i][j].be);
            cur.push_back(edg[i][j].en);
        }
        sort(cur.begin(), cur.end());
        cur.erase(unique(cur.begin(), cur.end()), cur.end());
        for (j = 0; j < cur.size(); j++)
        {
            int nummatches = 0;
            int q;
            int64 curplayer = cur[j];
            if (Find(curplayer)) continue;
            int cbeg = GetPos(curplayer);
            int alwayswin = 0;
            for (q = first[cbeg]; q != -1; q = next[q])
            {
                if (Find(edg[i][q].en)) continue;
                nummatches++;
                alwayswin = 1;
            }
            if (alwayswin == 1) continue;
            for (q = rfirst[cbeg]; q != -1; q = rnext[q])
            {
                if (Find(edg[i][q].be)) continue;
                nummatches++;
            }                  
            int64 lb, rb;
            rb = (curplayer >> i) << i;
            if ((curplayer & (numunder - 1)) != 0) rb += numunder;
            lb = rb - (numunder >> 1) + 1;
            if (curplayer >= lb)
            {
                lb -= numunder >> 1;
                rb -= numunder >> 1;
            }
            //cerr << "curplayer = " << curplayer << "\n";
            //cerr << "lb = " << lb << " rb = " << rb << "\n";
            //cerr << "bounds " << FindMinGE(lb) << " " << FindMaxLE(rb) << "\n";
            //cerr << "GetKilled = " << GetKilled(lb, rb) << "\n";
            int64 numgd = (rb - lb + 1) - (GetKilled(lb, rb));
            //cerr << "numgd = " << numgd << " nummatches = " << nummatches << "\n";
            if (nummatches == numgd)
            {
                killed[newnumkilled] = curplayer;
                newnumkilled++;
                maxtour[curplayer] = i;
            }
        }
        //cerr << "will be sorting\n";
        numkilled = newnumkilled;
        if (numkilled != 0)
        {
            sort(killed, killed + numkilled);
        }
        numunder <<= 1;
        //cerr << "Tour counted\n";
    }
    for (i = 0; i < res.size(); i++)
    {
        if (maxtour.find(res[i]) == maxtour.end())
        {
            printf("%d ", k);
        }
        else printf("%d ", maxtour[res[i]]);
    }
}

int main()
{
    freopen("topcoder.in", "rt", stdin);
    freopen("topcoder.out", "wt", stdout);
    srand(16567);
    Load();
    Solve();
    return 0;
}
